// project.e $Revision: 1.13 $ 
// 
// Fri, 09/16/1994 23:35:27 
// 
// Purpose: 
// 
//		This extension is designed as an aid to the individual who
//		has code or information within a project directory tree or
//		forest which is possibly widely scattered over the
//		development environment.  
// 
//		The primary purpose of this extension is to create an
//		environment in which all file operations (such as opens,
//		greps, etc. see details below) will work over all files in a
//		project.  The method by which this is accomplished is by
//		overloading the do_file_match routine.  The current values of
//		the new project variables are stored within the session files
//		saved during exit and restored on startup.  When files are
//		specified on the command line the previous session is
//		restored first and then the new files are added.  
// 
//		This extension defines two new variables which are set by the
//		user to specify the root directory location of the current
//		project and all directories which are a part of that project.
//		The project directory and its associated parts may all be
//		absolute or relative path names (which allows multiple
//		branches of a project on one machine at different absolute
//		locations but at similar relative locations for auxiliary
//		files).  I call the project a forest rather than simply a
//		tree because absolute files are allowed in the multiple
//		directory variable, project_directories, to allow the use of
//		one or more global locations at which files may be located
//		(ie \c700\include).  See the description of project_directory
//		and project_directories below to get more detail in setting
//		these variables.  
// 
//		NOTE: If an explicit file name is used then if there is more
//		than one file with the name specified, this extension
//		will return the file which occurs FIRST in the project list.
//
//		NOTE: This extension modifies the behavior of get_file and
//		get_file_dir completion routines to not return absolute file
//		names for completions with wildcards (ie. *.c, etc.)  Lugaru
//		had an exception for * and ? in file names, in all other
//		cases the Lugaru standard code also does this.  This is
//		required for proper dired behavior.
// 
//		NOTE: if the project_directories variable is non zero in
//		length then if the top level of the project should be
//		included in the enumeration, it must explicitly be a part of
//		project_directories.  
// 
// Usage: 
// 
//		This extension is very simple to configure.  Add project.doc
//		to the end of edoc, load project.b with the load-bytes
//		command and save the state using the command write-state.  
// 
//		The global project variables used by this extension
//		(project_directory and project_directories) are saved within
//		the session state file when exiting Epsilon and restored when
//		entering Epsilon.  This is accomplished via the session hooks
//		defined within session.e in the Epsilon distribution and do
//		not require that they be saved in the state file.  
// 
// 
// The modifications to this code are dedicated to the public domain 
// with the caveat that Lugaru is welcome to use this within their 
// distribution source code which is supplied with Epsilon.  Use it 
// any way you want for whatever purposes.  Portions are Copyright 
// Lugaru Software...  
// 
// 
// John Kercheval 
// 15563 10th Ave NE 
// Seattle, WA 98155-6209 
// INTERNET: johnk@wrq.com 
// Compu$erve: 72450,3702 
// 
// 
// Modification History: 
// 
//		Wed, 07/08/1992 23:44:04 V1.0 initial release.  
// 
//		Thu, 07/09/1992 22:43:46 V1.1 add session variables and
//		command line expansion support.  Adds significant usability
//		to the session mechanism and behaves much better for project
//		oriented and standard behavior edits.  Place the default
//		bindings for the replacement commands.  
// 
//		Thu, 07/09/1992 23:49:01 V1.2 add the commands
//		project_write_file, project_copy_to_file,
//		project_insert_file, project_write_region, load_file,
//		load_bytes and select_tag_file to cover the whole gambit of
//		routines using the get_file() routine.  
// 
//		Tue, 07/14/1992 21:59:13 V1.3 add read only check to session
//		restore hook so that files which have change read only status
//		will be correctly brought into the system.  
// 
//		Wed, 07/22/1992 21:02:26 V1.4 add more documentation to
//		header and add verbosity to the write_session command.  The
//		command find_project_file subroutine now works as does the
//		find_file routine in Epsilon with the except that completion
//		on dot directories behaves as does other directory
//		completions rather than giving an empty completion list.
//		Those not using the project features of this extension should
//		see behavior nearly identical with Epsilon's default.  Place
//		C-X C-F as project_find_file for compatibility reasons.
//		pluck_project_find_file is still a bindable command.  Find
//		file commands will open all files which fit the pattern and
//		show the last file found in the current buffer.  Fix bug in
//		absolute path check, break absolute path check into a
//		subroutine.  
// 
//		Thu, 07/23/1992 12:38:48 V1.5 write session will now save the
//		current session to the startup session file without
//		prompting.  If a numeric argument is given before the
//		write_session command then a file name prompt is given with
//		the default in the current project directory.  
// 
//		Mon, 07/27/1992 20:47:39 V1.6 fix bug opening new files using
//		project_find_file and project_pluck_file.  Use a unique
//		delimiter to fencepost the session information for the
//		project.  This will avoid collision with other session save
//		and restore hooks.  This is a dirty trick because it blows
//		away your previous project information, but this new approach
//		is relatively immune to competing session hooks and is order
//		independent.  
// 
//		Wed, 07/29/1992 21:34:36 V1.7 change session variables so
//		they all begin with session, this makes them easy to find via
//		completion.  Add new variable session_local_search to support
//		new directory hierarchy search for a project session file.
//		Fix a subtle bug in write_session and save_session which if
//		interrupted would store an invalid session_current_file
//		creating a problem on exit and storing the session in the
//		Epsilon default rather than in the startup project session.
//		Fix bug in read only buffer parsing init.  
// 
//		Tue, 09/08/1992 20:28:14 V1.8 remove the readonly file check
//		to reflect the fix performed by Lugaru for V6.02.  
// 
//		Thu, 09/24/1992 13:38:35 V1.9 write_session would save the
//		session always to the Epsilon default session file when it
//		should have been saving to the current active session file.  
// 
//		Sat, 02/06/1993 10:47:26 V2.0 remove session routines (except
//		for the session hooks), they are no longer needed.  Update
//		for V6.5 of Epsilon.  
// 
//		Sun, 10/31/1993 21:56:02 V2.1 beef up the project_pluck_file
//		routine to obtain full file specifications to match
//		<DRIVE>:<FILE_SPEC>(LINE:COLUMN) format.  Make the pluck code
//		load the file unless given an argument (makes it a bit more
//		useful).  Add the next-error file routines to aid compile
//		parsing.  
// 
//		Fri, 09/16/1994 23:35:47 V2.2 complete rewrite to take
//		advantage of V7.0 of Epsilon.  The method for hooking into
//		the system is simpler (an overload of do_file_match) 
// 
//		Thu, 11/10/1994 19:47:54 V2.3 move pluck code to it's own
//		module and correct a couple of Epsilon V7.0 B2 bugs (to be
//		removed later).
//
//		Mon, 01/23/1995 16:01:44 V2.4 Remove unneeded code related to
//		V7.0B2 bugs.
//
//		Tue, 05/23/1995 01:17:22 V2.5 Add support for project dired
//		(dired does not use the do_file_match() routine).  Correct
//		some logic problems in the file match code which have been
//		there for some time.  rearrange logic in file match code to
//		speed the file match up.
//
//		Tue, 05/30/1995 22:06:37 V2.6 Change the paradigm slightly to
//		not allow relative path searches within the project.  This
//		makes non project completion much more consistent.
//
//		Thu, 06/01/1995 00:00:52 V2.7 Cleaned up the completion
//		lister even more.  Correct a bug in Epsilon completion code
//		(requiring an overload of complete()) which should make
//		completion behaviour more consistent with 'normal' Epsilon
//		file completion behaviour.  Correct problem with exact match
//		completion during return (remove NONE_OK flag from completion
//		lister).
//
//		Thu, 06/01/1995 23:36:51 V2.8 Correct another problem with
//		complete.  NONE_OK is a required flag for some prompts
//		(notably grep).
//
// some amount of Epsilon original code so...
//
/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*   "EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.          *
*                                                                       *
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.  *
*                                                                       *
* Limited permission is hereby granted to reproduce and modify this     *
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in  *
* any such reproduction or modification.                                *
************************************************************************/

#include <eel.h>
#include <proc.h>
#include <tags.h>

#ifndef BOOL
#define BOOL int
#define TRUE 1
#define FALSE 0
#endif

#define MAX_BUF		256

#define PROJECT_START_DELIM	"<<SES!@!PROJECT_START>>"
#define PROJECT_END_DELIM	"<<SES!@!PROJECT_END>>"


//
// A list of subdirectories or absolute directories that project-grep
// will search in.  It is semicolon separated list just like PATH
// environment variable.  If a component is not absolute, it will be
// interpreted as relative to project_directory variable.  For example,
// suppose that project_directory is "c:\project" and this variable
// is set to "foo;d:\bar", then the following directories will be
// searched, in order (note the implicit first search of the current
// subdirectory),
//
//  the current directory
//  c:\project\foo
//  d:\bar
//
user char project_directories[MAX_BUF] = "";

//
// This is the root pathname for the project grep command.
// Directories mentioned in project_directories are interpreted as
// relative to this.  It may be absolute or relative, for example,
// "c:\project" or "project".  This variable should be relative if
// you have multiple branches for versions of a project on one system
// or you have multiple session files laying around.
//
user char project_directory[FNAMELEN] = "";


////////////////////////////////////////////////////////////////////////////
//
// is_absolute will return true if the passed path is absolute.
//

BOOL is_absolute_path(path)
	char *path;
{
	if (strlen(path) &&
		(path[0] == '\\' ||
		 path[0] == '/' ||
		 path[1] == ':')) {
		return TRUE;
	} else {
		return FALSE;
	}
}


////////////////////////////////////////////////////////////////////////////
//
// is_path will return true if the passed string is a path.  Note
// that a path may be absolute or relative.
//

BOOL is_path(path)
	char *path;
{
	if (index(path, '\\') ||
		index(path, '/') ||
		path[0] == '.') {
		return TRUE;
	} else {
		return FALSE;
	}
}


////////////////////////////////////////////////////////////////////////////
//
// maybe_prepend_path will combine the prefix and path if path is not
// absolute and return in ret, do not exceed length characters.
//

maybe_prepend_path(prefix, path, ret, length)
	char *prefix;				// path element to prepend
	char *path;					// relative path element
	char *ret;					// return string
	int length;					// maximum length of ret
{
	*ret = '\0';

	if (!strlen(path)) {
		strncpy(ret, prefix, length - 1);
	} else {
		if (is_absolute_path(path) ||
			!strlen(prefix)) {
			strncpy(ret, path, length - 1);
			return;
		} else {
			strncpy(ret, prefix, length - 1);
			if (path[strlen(path) - 1] == '\\' ||
				path[strlen(path) - 1] == '/' )
				strncat(ret, path, length - strlen(prefix) - 1);
			else {
				strncat(ret, "\\", length - strlen(prefix) - 1);
				strncat(ret, path, length - strlen(prefix) - 2);
			}
		}
	}
}


////////////////////////////////////////////////////////////////////////////
//
// maybe_append_path will combine the postfix and path and return in
// ret, do not exceed length characters.
//

maybe_append_path(postfix, path, ret, length)
	char *postfix;				// path element to append
	char *path;					// relative path element
	char *ret;					// return string
	int length;					// maximum length of ret
{
	*ret = '\0';

	if (!strlen(path)) {
		strncpy(ret, postfix, length - 1);
	} else {
		strncpy(ret, path, length - 1);
		if (path[strlen(path) - 1] == '\\' ||
			path[strlen(path) - 1] == '/' )
			strncat(ret, postfix, length - strlen(path) - 1);
		else {
			strncat(ret, "\\", length - strlen(path) - 1);
			strncat(ret, postfix, length - strlen(path) - 2);
		}
	}
}


////////////////////////////////////////////////////////////////////////////
//
// make_session_hook and restore_session_hook save the current
// project variables to the session file when it is written or read
// and then call any other defined session hooks.
//

new_project_make_session_hook(b)
{
	save_var bufnum = b;

	point = size();

	buf_printf(b, "\n%s", PROJECT_START_DELIM);
	buf_printf(b, "\n%s\n%s",
			   project_directory,
			   project_directories);
	buf_printf(b, "\n%s\n", PROJECT_END_DELIM);

	project_make_session_hook(b);
}

REPLACE_FUNC("project", "make_session_hook")

new_project_restore_session_hook(b)
{
	int old_point = point;

	save_var bufnum = b;

	//
	// obtain the project variables
	//
	point = 0;
	if (!search(1, PROJECT_START_DELIM)) {
		point = old_point;
	} else {
		nl_forward();			 // move to first variable
		grab_line(b, project_directory);
		grab_line(b, project_directories);
		nl_forward();			 // move beyond final delimiter
	}

	//
	// call the previous hook
	//
	project_restore_session_hook(b);
}

REPLACE_FUNC("project", "restore_session_hook")


////////////////////////////////////////////////////////////////////////////
//
// get_project_directory will return all project subdirectories one
// at a time and return FALSE when no more subdirectories are found.
// Flags is set to MATCHSTART when the first subdirectory is required
// (reset).  The return string will not exceed length characters.
//

//
// This is an index used during the parsing of the project path
//
int project_directory_index;

BOOL
get_project_directory(ret, length, flags)
	char *ret;					// return string
	int length;					// maximum length of return string
	int flags;					// standard completion flags
{
	char *begin_directory;		// the start of this element
	char *end_directory;		// the end of this element
	char ret_root[FNAMELEN];	// the intermediate return value
	//
	// Are we just starting the match?  If the project_directories
	// variable is of zero length then the directory specified by the
	// project_directory variable is the only element.  NOTE: if the
	// project_directories variable is non zero in length then if the
	// top level of the project should be included in the
	// enumeration, it must explicitly be a part of
	// project_directories.
	//
	if (flags & STARTMATCH) {
		project_directory_index = 0;
		*ret = '\0';
	}
	if (flags & STARTMATCH &&
		!project_directories[0]) {
		return TRUE;
	}
	//
	// If we are at the end of the list return FALSE
	//
	if (project_directory_index >= strlen(project_directories) - 1) {
		*ret = '\0';
		return FALSE;
	}
	//
	// Move over leading semicolons and place at beginning of
	// possible match
	//
	begin_directory = project_directories + project_directory_index;
	while (*begin_directory == ';') {
		begin_directory++;
	}
	//
	// If at the end of the string we are done
	//
	if (*begin_directory == '\0') {
		project_directory_index = begin_directory - project_directories;
		*ret = '\0';
		return FALSE;
	}
	//
	// We found something, copy it over and adjust statics accordingly
	//
	if ((end_directory = strstr(begin_directory, ";")) == NULL) {
		strncpy(ret_root, begin_directory, length - 1);
		project_directory_index = strlen(project_directories);
	} else {
		if (end_directory - begin_directory < length) {
			strncpy(ret_root, begin_directory,
					end_directory - begin_directory);
		} else {
			strncpy(ret_root, begin_directory, length - 1);
		}
		project_directory_index = end_directory - project_directories;
	}
	maybe_prepend_path(project_directory, ret_root, ret, length);

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////
//
// new_project_do_file_match will return all project files which match
// partial one at a time and return NULL when no more files are
// found.  Flags is set to FM_PREFIX, FM_EXACT when the first match
// is required (reset).  The return value will include the project
// directory information.
//

//
// This is an intermediate path variable used to track the completion
// state for the project file lister and match routine.
//
char project_current_directory[FNAMELEN] = "";

//
// This is the initial startup flags used.  This is stored for reuse
// as a parameter to file_match on successive project directories.
//
int start_flags = 0;

char *
new_project_do_file_match(partial, flags)
	char *partial;				// return string
	int flags;					// standard completion flags
{
	BOOL found = FALSE;			// sentinal
	BOOL start = FALSE;			// file_match start state

	char project_current_file[FNAMELEN];	// full completion string
	char *ret;					// return value

	//
	// if there is no project directory or if we are expanding an
	// absolute or relative path, then perform standard behavior.
	//
	if ((!project_directory[0] &&
		 !project_directories[0]) ||
		is_path(partial)) {
		return file_match(partial, flags);
	}

	//
	// Init
	//
	if (flags & FM_EXACT ||
		flags & FM_PREFIX) {
		get_project_directory(project_current_directory,
							  FNAMELEN, STARTMATCH);
		start_flags = flags;
		start = TRUE;
	}

	//
	// cycle through the project list
	//
	while (!found) {
		maybe_append_path(partial, project_current_directory,
						  project_current_file, FNAMELEN);
		//
		// If a match is forthcoming then leave the loop, otherwise
		// enumerate the project directory and try again.
		//
		if ((ret = file_match(project_current_file,
							  start ? start_flags : 0)) != NULL) {
			found = TRUE;
		} else if (!get_project_directory(project_current_directory,
										  FNAMELEN, 0)) {
			start_flags = 0;
			return NULL;
		} else {
			start = TRUE;
		}
	}

	if (!ret) {
		start_flags = 0;
	}
	return ret;
}

REPLACE_FUNC("project", "do_file_match")


////////////////////////////////////////////////////////////////////////////
//
// new_project_get_file overloads get_file and provides more
// restricted completion when the project module is in use.
//

int list_file_matches();

new_project_get_file(res, pr, def)
	char *res, *pr, *def;
{
	char rel[FNAMELEN];
	int flags = CAUTIOUS | FOLDFILE;

	relative(def, rel);
	save_var completion_lister = list_file_matches;
	save_var list_finder = f_all_match;
	save_var abort_file_matching = ABORT_JUMP;

	if (project_directory[0] ||
		project_directories[0]) {
		//
		// for project files, assume that file creation is rare.
		//
		flags |= MUST_MATCH | NONE_OK;
	}

	prompt_comp_read(res, pr, f_match, flags, rel);
	if (*res && !wild_pattern(res))
		absolute(res);
}

REPLACE_FUNC("project", "get_file")


////////////////////////////////////////////////////////////////////////////
//
// new_project_get_file_dir overloads get_file_dir and provides more
// restricted completion when the project module is in use.
//

new_project_get_file_dir(res, pr)
	char *res, *pr;
{
	char def[FNAMELEN], *p;
	int flags = CAUTIOUS | FOLDFILE;

	if (is_dired_buf())		/* get directory name from dired */
		dired_getdir(def);
	else {
		strcpy(def, filename);
		absolute(def);
		if (*filename)	/* want only directory part */
			*get_tail(def, 1) = 0;
		p = def + strlen(def);	/* ptr to end of string */
		if (!is_path_separator(p[-1])) {
			*p++ = path_sep; /* make sure it ends with slash */
			*p = 0;
		}
	}
	save_var completion_lister = list_file_matches;
	save_var list_finder = f_all_match;
	save_var abort_file_matching = ABORT_JUMP;

	if (project_directory[0] ||
		project_directories[0]) {
		//
		// for project files, assume that file creation is rare.
		//
		flags |= MUST_MATCH | NONE_OK;
	}

	comp_read(res, pr, f_match, flags, def);

	if (*res && !wild_pattern(res))
		absolute(res);
}

REPLACE_FUNC("project", "get_file_dir")


////////////////////////////////////////////////////////////////////////////
//
// new_project_complete overloads complete to correct assumptions
// about the type of data being returned when using the cautious
// flag. 
//

new_project_complete(s, finder, flags)	/* perform completion by modifying s */
	char *s;		/* return 0: no matches, 1: 1 match, */
	char *(*finder)();	/* else many matches, or some matches */
{				/* don't start with s */
	char biggest[FNAMELEN], *t;
	int i, start, end, num = 1;

	if (!(t = (*finder)(s, STARTMATCH)))		/* if no matches */
		return 0;
	strcpy(biggest, t);			/* start off with first */
	start = (flags & CAUTIOUS) ? 0 : strlen(s);
			/* assume all matches start with s unless cautious */
	save_var case_fold = (flags & COMP_FOLD);
	end = strlen(biggest);
	while (t = (*finder)(s, 0)) {
		if (strnfcmp(t, biggest, end))	/* if this differs, */
			for (i = start; i < end; i++) /* find where */
				if (charfcmp(t[i], biggest[i]))
					end = i;
		num++;
		if (end <= start)	/* if all chars gone, stop */
			break;
	}
	if (flags & CAUTIOUS &&
		end < start) {
			return 2; // some matches didn't start with s
	}

	if (end > strlen(s)) {
		strncpy(s, biggest, end);
	}
	
	return num;
}

REPLACE_FUNC("project", "complete")

	
////////////////////////////////////////////////////////////////////////////
//
// new_project_must_match overloads must_match to correct exact
// matches when expanding project directories with subdirectory
// wildcards. 
//

new_project_must_match(trystr, finder, flags)	/* return 1 if we should put up */
char *trystr;				/* list of possible completions */
char *(*finder)();
{
	char first[FNAMELEN], *s;

	if (!(flags & MUST_MATCH) || (flags & NONE_OK) && !*trystr)
		return 0;
	
	s = (*finder)(trystr, STARTMATCH);
	if (!s)			/* no matches at all, return it */
		return 0;
	
	strcpy(first, s);
	if (!(*finder)(trystr, 0)) {
		strcpy(trystr, first);	/* 1 match, expand it */
		return 0;
	} else if (wild_pattern(trystr)) {
		/* two or more matches but wild cards are dired candidates */
			return 0;
	} else {
		/* two or more matches, ask */
		return 1;
	}
}

REPLACE_FUNC("project", "must_match")

	
////////////////////////////////////////////////////////////////////////////
//
// new_project_make_dired overloads make_dired and loads all files in
// the project matching dir in to the current dired buffer.
//

new_project_make_dired(dir)
char *dir;
{
	char project_current_path[FNAMELEN];	// full completion string
	int ret = 0;

	modified = 0;
	save_var abort_file_matching = ABORT_JUMP;

	//
	// If we have no project variable or are expanding an absolute or
	// relative path don't bother with the project expansions.
	//
	if ((!project_directory[0] &&
		 !project_directories[0]) ||
		is_path(dir)) {
		if (do_dired(dir)) {
			ret++;
		}
	} else {
		//
		// cycle through all of the project directories.
		//
		get_project_directory(project_current_directory,
							  FNAMELEN, STARTMATCH);
		do {
			maybe_append_path(dir, project_current_directory,
							  project_current_path, FNAMELEN);
			if (do_dired(project_current_path)) {
				ret++;
			}
		} while (get_project_directory(project_current_directory,
									   FNAMELEN, 0));
	}

	modified = 0;
	if (ret) {
		sort_for_dired();
	}
	return ret;
}


REPLACE_FUNC("project", "make_dired")


////////////////////////////////////////////////////////////////////////////
//
// new_project_wild_pattern overloads wild_pattern and returns true
// if any wild card character is used in the file specification.
//

new_project_wild_pattern(s)	/* does this filename use wildcards? */
char *s;
{
	return (index(s, '{') ||
			index(s, '}') ||
			index(s, ',') ||
			index(s, ';') ||
			index(s, '*') ||
			index(s, '?'));
}

REPLACE_FUNC("project", "wild_pattern")

